import numpy as np
from mot.cl_routines.mapping.run_procedure import RunProcedure
from ...utils import KernelInputArray, SimpleNamedCLFunction, KernelInputLocalMemory, KernelInputAllocatedOutput
from ...cl_routines.base import CLRoutine


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ObjectiveFunctionCalculator(CLRoutine):

    def calculate(self, model, parameters):
        """Calculate and return the objective function of the given model for the given parameters.

        Args:
            model (AbstractModel): The model to calculate the objective function for
            parameters (ndarray): The parameters to use in the evaluation of the model, an (d, p) matrix
                with d problems and p parameters.

        Returns:
            ndarray: per problem the objective function.
        """
        shape = parameters.shape
        all_kernel_data = dict(model.get_kernel_data())
        all_kernel_data.update({
            'parameters': KernelInputArray(parameters),
            'objective_values': KernelInputAllocatedOutput((shape[0],), 'mot_float_type'),
            'local_reduction_lls': KernelInputLocalMemory('double')
        })

        runner = RunProcedure(**self.get_cl_routine_kwargs())
        runner.run_procedure(self._get_wrapped_function(model, parameters), all_kernel_data, parameters.shape[0],
                             double_precision=model.double_precision, use_local_reduction=True)

        return all_kernel_data['objective_values'].get_data()

    def _get_wrapped_function(self, model, parameters):
        objective_func = model.get_objective_per_observation_function()
        nmr_params = parameters.shape[1]

        func = ''
        func += objective_func.get_cl_code()
        func += '''
            void _fill_objective_value_tmp(mot_data_struct* data,
                                          mot_float_type* x,
                                          local double* objective_value_tmp){

                ulong observation_ind;
                ulong local_id = get_local_id(0);
                objective_value_tmp[local_id] = 0;
                uint workgroup_size = get_local_size(0);
                uint elements_for_workitem = ceil(''' + str(model.get_nmr_inst_per_problem()) + ''' 
                                                  / (mot_float_type)workgroup_size);
                
                if(workgroup_size * (elements_for_workitem - 1) + local_id 
                        >= ''' + str(model.get_nmr_inst_per_problem()) + '''){
                    elements_for_workitem -= 1;
                }
                
                for(uint i = 0; i < elements_for_workitem; i++){
                    observation_ind = i * workgroup_size + local_id;
                    objective_value_tmp[local_id] += ''' + objective_func.get_cl_function_name() + '''(
                        data, x, observation_ind);
                }

                barrier(CLK_LOCAL_MEM_FENCE);
            }

            double _sum_objective_value_tmp(local double* objective_value_tmp){
                double ll = 0;
                for(uint i = 0; i < get_local_size(0); i++){
                    ll += objective_value_tmp[i];
                }
                return ll;
            }
            
        '''
        func += '''
            void compute(mot_data_struct* data){
                mot_float_type x[''' + str(nmr_params) + '''];
                for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                    x[i] = data->parameters[i];
                }
                
                _fill_objective_value_tmp(data, x, data->local_reduction_lls);
                if(get_local_id(0) == 0){
                    *(data->objective_values) = _sum_objective_value_tmp(data->local_reduction_lls);
                }
            }
        '''
        return SimpleNamedCLFunction(func, 'compute')



import os
from pkg_resources import resource_filename

from mot.library_functions import LibNMSimplex
from ...cl_routines.optimizing.base import AbstractParallelOptimizer, AbstractParallelOptimizerWorker

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SBPlex(AbstractParallelOptimizer):

    default_patience = 5

    def __init__(self, scale=None, alpha=None, beta=None, gamma=None, delta=None, psi=None, omega=None,
                 min_subspace_length=None, max_subspace_length=None, adaptive_scales=None,
                 patience_nmsimplex=None,
                 patience=None, optimizer_settings=None, **kwargs):
        """Variation on the Nelder-Mead Simplex method by Thomas H. Rowan.

        This method uses NMSimplex to search subspace regions for the minimum. See Rowan's thesis titled
        "Functional Stability analysis of numerical algorithms" for more details.

        Args:
            patience (int): Used to set the maximum number of iterations to patience*(number_of_parameters+1)
            patience_nmsimplex (int): The maximum patience for each subspace search.
                For each subspace search we set the number of iterations to patience*(number_of_parameters_subspace+1)
            scale (double): the scale of the initial simplex, default 1.0
            alpha (double): reflection coefficient, default 1.0
            beta (double): contraction coefficient, default 0.5
            gamma (double); expansion coefficient, default 2.0
            delta (double); shrinkage coefficient, default 0.5
            psi (double): subplex specific, simplex reduction coefficient, default 0.01
                The default used in Rowan's Thesis is 0.25, we opted here for 0.01 for greater accuracy.
            omega (double): subplex specific, scaling reduction coefficient, default 0.1
            min_subspace_length (int): the minimum subspace length, defaults to min(2, n).
                This should hold: (1 <= min_s_d <= max_s_d <= n and min_s_d*ceil(n/nsmax_s_dmax) <= n)
            max_subspace_length (int): the maximum subspace length, defaults to min(5, n).
                This should hold: (1 <= min_s_d <= max_s_d <= n and min_s_d*ceil(n/max_s_d) <= n)

            adaptive_scales (boolean): if set to True we use adaptive scales instead of the default scale values.
                This sets the scales to:

                .. code-block:: python

                    n = model.get_nmr_estimable_parameters()

                    alpha = 1
                    beta  = 0.75 - 1.0 / (2 * n)
                    gamma = 1 + 2.0 / n
                    delta = 1 - 1.0 / n

                Following the paper:

                * Gao F, Han L. Implementing the Nelder-Mead simplex algorithm with adaptive parameters.
                  Comput Optim Appl. 2012;51(1):259-277. doi:10.1007/s10589-010-9329-3.

            optimizer_settings (dict): the optimization settings, you can use the following:
                scale, alpha, beta, gamma, delta, adaptive_scales

                The scales should satisfy the following constraints:

                .. code-block:: python

                    alpha > 0
                    0 < beta < 1
                    gamma > 1
                    gamma > alpha
                    0 < delta < 1
        """
        patience = patience or self.default_patience

        optimizer_settings = optimizer_settings or {}

        keyword_values = {}
        keyword_values['scale'] = scale
        keyword_values['alpha'] = alpha
        keyword_values['beta'] = beta
        keyword_values['gamma'] = gamma
        keyword_values['delta'] = delta
        keyword_values['adaptive_scales'] = adaptive_scales
        keyword_values['psi'] = psi
        keyword_values['omega'] = omega
        keyword_values['min_subspace_length'] = min_subspace_length
        keyword_values['max_subspace_length'] = max_subspace_length
        keyword_values['patience_nmsimplex'] = patience_nmsimplex

        option_defaults = {'alpha': 1.0, 'beta': 0.5, 'gamma': 2.0, 'delta': 0.5, 'scale': 1.0,
                           'adaptive_scales': True, 'psi': 0.01, 'omega': 0.1, 'min_subspace_length': 'auto',
                           'max_subspace_length': 'auto', 'patience_nmsimplex': 10}

        def get_value(option_name):
            value = keyword_values.get(option_name)
            if value is None:
                value = optimizer_settings.get(option_name)
            if value is None:
                value = option_defaults[option_name]
            return value

        for option in option_defaults:
            optimizer_settings.update({option: get_value(option)})

        super(SBPlex, self).__init__(patience=patience, optimizer_settings=optimizer_settings, **kwargs)

    def minimize(self, model, starting_positions):
        nmr_params = model.get_nmr_estimable_parameters()

        if self._optimizer_settings.get('adaptive_scales', True):
            self._optimizer_settings.update(
                {'alpha': 1,
                 'beta': 0.75 - 1.0 / (2 * nmr_params),
                 'gamma': 1 + 2.0 / nmr_params,
                 'delta': 1 - 1.0 / nmr_params}
                )

        if self._optimizer_settings.get('min_subspace_length', 'auto') == 'auto':
            self._optimizer_settings.update({'min_subspace_length': min(2, nmr_params)})

        if self._optimizer_settings.get('max_subspace_length', 'auto') == 'auto':
            self._optimizer_settings.update({'max_subspace_length': min(5, nmr_params)})

        return super(SBPlex, self).minimize(model, starting_positions)

    def _get_worker_generator(self, *args):
        return lambda cl_environment: SBPlexWorker(cl_environment, *args)


class SBPlexWorker(AbstractParallelOptimizerWorker):

    def _get_optimization_function(self):
        params = {'NMR_PARAMS': self._nmr_params, 'PATIENCE': self._parent_optimizer.patience}

        for option, value in self._optimizer_settings.items():
            if option == 'scale':
                params['INITIAL_SIMPLEX_SCALES'] = '{' + ', '.join([str(value)] * self._nmr_params) + '}'
            else:
                params.update({option.upper(): value})

        sbplex = open(os.path.abspath(resource_filename('mot', 'data/opencl/sbplex.cl')), 'r').read()
        if params:
            sbplex = sbplex % params

        body = sbplex
        body += LibNMSimplex(evaluate_fname='subspace_evaluate').get_cl_code()

        return body

    def _get_optimizer_call_name(self):
        return 'sbplex'



import pyopencl as cl
import numpy as np
from ...utils import get_float_type_def, all_logging_disabled, KernelInputDataManager
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker
from copy import copy

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class WAICCalculator(CLRoutine):

    def calculate(self, model, samples):
        r"""Compute the Watanabe-Akaike information criterion (WAIC; Watanabe, 2010) for the given samples.

        See http://www.stat.columbia.edu/~gelman/research/unpublished/waic_stan.pdf for a gentle introduction to
        this implementation.

        The WAIC can be used as a model selection criteria, in which the model with the lowest WAIC is preferred.

        This is an expensive function to call, considering that it needs to calculate the log likelihoods for every
        observation separately.

        Args:
            model (AbstractModel): The model to calculate the WAIC for
            samples (ndarray): The obtained samples per problem. This is supposed to be a matrix
                of shape (d, p, s) with d problems, p parameters and s samples.

        Returns:
            ndarray: per problem the calculated WAIC
        """
        mot_float_dtype = np.float32
        if model.double_precision:
            mot_float_dtype = np.float64

        nmr_problems = samples.shape[0]
        nmr_observations = model.get_nmr_inst_per_problem()

        waics = np.zeros((nmr_problems,))

        samples = np.require(samples, dtype=mot_float_dtype, requirements=['C', 'A', 'O'])
        logsumexps = np.zeros(nmr_observations, dtype=np.float64, order='C')
        variances = np.zeros(nmr_observations, dtype=np.float64, order='C')

        workers = self._create_workers(
            lambda cl_environment: _LLWorker(cl_environment,
                                             self.get_compile_flags_list(model.double_precision),
                                             model, samples, logsumexps, variances, mot_float_dtype))

        for problem_ind in range(nmr_problems):

            for worker in workers:
                worker.set_problem_index(problem_ind)

            with all_logging_disabled():
                self.load_balancer.process(workers, nmr_observations)

            lpd = np.sum(logsumexps)
            p_waic = np.sum(variances)

            waics[problem_ind] = -2 * (lpd - p_waic)

        return waics


class _LLWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, samples, logsumexps, variances, mot_float_dtype):
        super(_LLWorker, self).__init__(cl_environment)

        self._problem_index = 0

        self._model = model
        self._data_info = self._model.get_kernel_data()
        self._data_struct_manager = KernelInputDataManager(self._data_info, mot_float_dtype)
        self._double_precision = model.double_precision

        self._samples = samples
        self._samples_per_voxel = self._samples[self._problem_index]
        self._log_sum_exps = logsumexps
        self._variances = variances

        self._nmr_samples = self._samples.shape[2]
        self._nmr_params = self._samples.shape[1]
        self._nmr_observations = self._log_sum_exps.shape[0]

        self._log_likelihoods = np.zeros([self._nmr_observations, self._nmr_samples], dtype=np.float64, order='C')

        self._ll_calculating_buffers, self._ll_buffer, self._lse_buffer, self._variances_buffer = self._create_buffers()

        self._ll_calculating_kernel = self._build_kernel(self._get_ll_calculating_kernel(), compile_flags)
        self._statistic_kernels = self._build_kernel(self._get_statistics_kernels(), compile_flags)

    def set_problem_index(self, problem_index):
        """Set the problem index of the samples we are currently working on."""
        self._problem_index = problem_index
        samples = self._samples[self._problem_index]
        mapped_buf = cl.enqueue_map_buffer(
            self._cl_queue, self._ll_calculating_buffers[0],
            cl.map_flags.WRITE, 0, samples.shape, samples.dtype)[0]
        mapped_buf[:] = samples

    def calculate(self, range_start, range_end):
        self._calculate_lls(range_start, range_end)
        self._calculate_statistics(range_start, range_end)

        self._enqueue_readout(self._lse_buffer, self._log_sum_exps, range_start, range_end)
        self._enqueue_readout(self._variances_buffer, self._variances, range_start, range_end)

    def _calculate_lls(self, range_start, range_end):
        nmr_problems = range_end - range_start

        buffers = copy(self._ll_calculating_buffers)
        buffers.insert(0, np.uint32(self._problem_index))

        arg_dtypes = [np.uint32, None, None]
        arg_dtypes.extend(self._data_struct_manager.get_scalar_arg_dtypes())

        kernel = self._ll_calculating_kernel.run_kernel
        kernel.set_scalar_arg_dtypes(arg_dtypes)

        kernel(self._cl_queue, (int(nmr_problems), int(self._nmr_samples)), None,
               *buffers, global_offset=(int(range_start), 0))

    def _calculate_statistics(self, range_start, range_end):
        nmr_problems = range_end - range_start

        max_work_group_sizes = [
            cl.Kernel(self._statistic_kernels, 'mean_and_max').get_work_group_info(
                cl.kernel_work_group_info.WORK_GROUP_SIZE, self._cl_environment.device),
            cl.Kernel(self._statistic_kernels, 'logsum_variance').get_work_group_info(
                cl.kernel_work_group_info.WORK_GROUP_SIZE, self._cl_environment.device)
        ]
        workgroup_size = min(max_work_group_sizes)

        lse_tmp_buffer = cl.LocalMemory(workgroup_size * np.dtype('double').itemsize)
        var_tmp_buffer = cl.LocalMemory(workgroup_size * np.dtype('double').itemsize)

        buffers = [self._ll_buffer, self._lse_buffer, self._variances_buffer, lse_tmp_buffer, var_tmp_buffer]

        self._statistic_kernels.mean_and_max(
            self._cl_queue, (int(nmr_problems * workgroup_size),), (int(workgroup_size),),
            *buffers, global_offset=(int(range_start * workgroup_size),))

        self._statistic_kernels.logsum_variance(
            self._cl_queue, (int(nmr_problems * workgroup_size),), (int(workgroup_size),),
            *buffers, global_offset=(int(range_start * workgroup_size),))

    def _create_buffers(self):
        ll_buffer = cl.Buffer(self._cl_context,
                              cl.mem_flags.READ_WRITE,
                              size=(np.dtype(np.float64).itemsize * self._nmr_observations * self._nmr_samples))

        lse_buffer = cl.Buffer(self._cl_context,
                               cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                               hostbuf=self._log_sum_exps)

        variances_buffer = cl.Buffer(self._cl_context,
                                     cl.mem_flags.READ_WRITE | cl.mem_flags.USE_HOST_PTR,
                                     hostbuf=self._variances)

        samples_buffer = cl.Buffer(self._cl_context,
                                   cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                   hostbuf=self._samples_per_voxel)

        ll_calculating_buffers = [samples_buffer, ll_buffer]
        ll_calculating_buffers.extend(self._data_struct_manager.get_kernel_inputs(self._cl_context, 1))

        return ll_calculating_buffers, ll_buffer, lse_buffer, variances_buffer

    def _get_ll_calculating_kernel(self):
        """Kernel to calculate the log likelihoods per sample per observation"""
        obs_func = self._model.get_log_likelihood_per_observation_function()

        kernel_param_names = ['uint problem_ind', 'global mot_float_type* samples', 'global double* lls']
        kernel_param_names.extend(self._data_struct_manager.get_kernel_arguments())
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += self._data_struct_manager.get_struct_definition()
        kernel_source += obs_func.get_cl_code()
        kernel_source += r'''
            __kernel void run_kernel(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    ulong obs_ind = get_global_id(0);
                    ulong sample_ind = get_global_id(1);

                    mot_data_struct data = ''' + self._data_struct_manager.get_struct_init_string('problem_ind') + ''';
                    mot_float_type x[''' + str(self._nmr_params) + '''];

                    for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        x[i] = samples[i * ''' + str(self._nmr_samples) + ''' + sample_ind];
                    }

                    lls[obs_ind * ''' + str(self._nmr_samples) + ''' + sample_ind] =
                        ''' + obs_func.get_cl_function_name() + '''(&data, x, obs_ind);
                }
        '''
        return kernel_source

    def _get_statistics_kernels(self):
        kernel_param_names = ['global double* lls', 'global double* lses', 'global double* variances',
                              'local double* lse_tmp', 'local double* var_tmp']
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += r'''
            __kernel void mean_and_max(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    ulong obs_ind = (ulong)(get_global_id(0) / get_local_size(0));
                    ulong local_id = get_local_id(0);
                    uint workgroup_size = get_local_size(0);
                    uint elements_for_workitem = ceil(''' + str(self._nmr_samples) + ''' 
                                                      / (mot_float_type)workgroup_size);
                
                    if(workgroup_size * (elements_for_workitem - 1) + local_id >= ''' + str(self._nmr_samples) + '''){
                        elements_for_workitem -= 1;
                    }
                
                    ulong sample_ind;

                    double ll;
                    double max_ll = -HUGE_VAL;
                    double mean_sum = 0;

                    for(uint i = 0; i < elements_for_workitem; i++){
                        sample_ind = i * workgroup_size + local_id;
                        
                        ll = lls[obs_ind * ''' + str(self._nmr_samples) + ''' + sample_ind];

                        max_ll = max(max_ll, ll);
                        mean_sum += ll;
                    }
                    lse_tmp[local_id] = max_ll;
                    var_tmp[local_id] = mean_sum;

                    barrier(CLK_LOCAL_MEM_FENCE);

                    if(local_id == 0){
                        mean_sum = 0;
                        for(uint i = 0; i < workgroup_size; i++){
                            max_ll = max(max_ll, lse_tmp[i]);
                            mean_sum += var_tmp[i];
                        }
                        lses[obs_ind] = max_ll;
                        variances[obs_ind] = mean_sum / ''' + str(self._nmr_samples) + ''';
                    }

                }

            __kernel void logsum_variance(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    ulong obs_ind = (ulong)(get_global_id(0) / get_local_size(0));
                    ulong sample_ind;
                    ulong local_id = get_local_id(0);
                    uint workgroup_size = get_local_size(0);

                    double ll;
                    double max_ll = lses[obs_ind];
                    double mean_ll = variances[obs_ind];
                    double exp_sum = 0;
                    double var_sum = 0;

                    for(uint i = 0; i < ceil(''' + str(self._nmr_samples) + ''' / (mot_float_type)workgroup_size); i++){
                        sample_ind = i * workgroup_size + local_id;

                        if(sample_ind < ''' + str(self._nmr_samples) + '''){
                            ll = lls[obs_ind * ''' + str(self._nmr_samples) + ''' + sample_ind];

                            exp_sum += exp(ll - max_ll);
                            var_sum += pown(ll - mean_ll, 2);
                        }
                    }
                    lse_tmp[local_id] = exp_sum;
                    var_tmp[local_id] = var_sum;

                    barrier(CLK_LOCAL_MEM_FENCE);

                    if(local_id == 0){
                        exp_sum = 0;
                        var_sum = 0;
                        for(uint i = 0; i < workgroup_size; i++){
                            exp_sum += lse_tmp[i];
                            var_sum += var_tmp[i];
                        }

                        lses[obs_ind] += log(exp_sum / ''' + str(self._nmr_samples) + ''');
                        variances[obs_ind] = var_sum / (''' + str(self._nmr_samples) + ''' - 1);
                    }
                }
        '''
        return kernel_source



import numpy as np

from mot.cl_routines.mapping.run_procedure import RunProcedure
from ...utils import SimpleNamedCLFunction, KernelInputArray, KernelInputData, is_scalar, \
    KernelInputScalar, KernelInputAllocatedOutput
from ...cl_routines.base import CLRoutine


__author__ = 'Robbert Harms'
__date__ = '2017-08-31'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class CLFunctionEvaluator(CLRoutine):

    def __init__(self, **kwargs):
        """This class can evaluate an arbitrary CL function implementation on some input data.
        """
        super(CLFunctionEvaluator, self).__init__(**kwargs)

    def evaluate(self, cl_function, input_data, double_precision=False, return_inputs=False):
        """Evaluate the given CL function at the given data points.

        This function will convert possible dots in the parameter name to underscores for in the CL kernel.

        Args:
            cl_function (mot.cl_function.CLFunction): the CL function to evaluate
            input_data (dict): for each parameter of the function either an array with input data or an
                :class:`mot.utils.KernelInputData` object. Each of these input datasets must either be a scalar or be
                of equal length in the first dimension. The user can either input raw ndarrays or input
                KernelInputData objects. If an ndarray is given we will load it read/write by default.
            double_precision (boolean): if the function should be evaluated in double precision or not
            return_inputs (boolean): if we are interested in the values of the input arrays after evaluation.

        Returns:
            ndarray or tuple(ndarray, dict[str: ndarray]): we always return at least the return values of the function,
                which can be None if this function has a void return type. If ``return_inputs`` is set to True then
                we return a tuple with as first element the return value and as second element a dictionary mapping
                the output state of the parameters.
        """
        nmr_data_points = self._get_minimum_data_length(input_data)

        kernel_items = self._wrap_input_data(cl_function, input_data, double_precision)

        if cl_function.get_return_type() != 'void':
            kernel_items['_results'] = KernelInputAllocatedOutput((nmr_data_points,), cl_function.get_return_type())

        runner = RunProcedure(**self.get_cl_routine_kwargs())
        runner.run_procedure(self._wrap_cl_function(cl_function, kernel_items),
                             kernel_items, nmr_data_points, double_precision=double_precision)

        if cl_function.get_return_type() != 'void':
            return_value = kernel_items['_results'].get_data()
            del kernel_items['_results']
        else:
            return_value = None

        if return_inputs:
            return return_value, {key: value.get_data() for key, value in kernel_items.items()}
        return return_value

    def _wrap_input_data(self, cl_function, input_data, double_precision):
        min_data_length = self._get_minimum_data_length(input_data)

        kernel_items = {}
        for param in cl_function.get_parameters():
            if isinstance(input_data[param.name], KernelInputData):
                kernel_items[self._get_param_cl_name(param.name)] = input_data[param.name]
            elif is_scalar(input_data[param.name]) and not param.data_type.is_pointer_type:
                kernel_items[self._get_param_cl_name(param.name)] = KernelInputScalar(input_data[param.name])
            else:
                if is_scalar(input_data[param.name]):
                    data = np.ones(min_data_length) * input_data[param.name]
                else:
                    data = input_data[param.name]

                kernel_items[self._get_param_cl_name(param.name)] = KernelInputArray(
                    data, ctype=param.data_type.ctype, is_writable=True, is_readable=True)

        return kernel_items

    def _get_minimum_data_length(self, input_data):
        min_length = 1

        for value in input_data.values():
            if isinstance(value, KernelInputData):
                data = value.get_data()
                if data is not None:
                    if np.ndarray(data).shape[0] > min_length:
                        min_length = np.ndarray(data).shape[0]
            elif is_scalar(value):
                pass
            elif value.shape[0] > min_length:
                min_length = value.shape[0]

        return min_length

    def _wrap_cl_function(self, cl_function, kernel_items):
        func_args = []
        for param in cl_function.get_parameters():
            param_cl_name = self._get_param_cl_name(param.name)

            if kernel_items[param_cl_name].is_scalar:
                func_args.append('data->{}'.format(param_cl_name))
            else:
                if param.data_type.is_pointer_type:
                    func_args.append('data->{}'.format(param_cl_name))
                else:
                    func_args.append('data->{}[0]'.format(param_cl_name))

        func_name = 'evaluate'
        func = cl_function.get_cl_code()

        if cl_function.get_return_type() == 'void':
            func += '''
                void ''' + func_name + '''(mot_data_struct* data){
                    ''' + cl_function.get_cl_function_name() + '''(''' + ', '.join(func_args) + ''');  
                }
            '''
        else:
            func += '''
                void ''' + func_name + '''(mot_data_struct* data){
                    *(data->_results) = ''' + cl_function.get_cl_function_name() + '(' + ', '.join(func_args) + ''');  
                }
            '''
        return SimpleNamedCLFunction(func, func_name)

    def _get_param_cl_name(self, param_name):
        if '.' in param_name:
            return param_name.replace('.', '_')
        return param_name



import os
from pkg_resources import resource_filename
import pyopencl as cl
from mot.utils import get_float_type_def
from .base import AbstractParallelOptimizer, AbstractParallelOptimizerWorker

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class LevenbergMarquardt(AbstractParallelOptimizer):

    default_patience = 250

    def __init__(self, patience=None, step_bound=None, scale_diag=None, optimizer_settings=None, **kwargs):
        """Use the Levenberg-Marquardt method to calculate the optimimum.

        Args:
            patience (int): Used to set the maximum number of iterations to patience*(number_of_parameters+1)
        """
        patience = patience or self.default_patience

        optimizer_settings = optimizer_settings or {}

        keyword_values = {}
        keyword_values['step_bound'] = step_bound
        keyword_values['scale_diag'] = scale_diag

        option_defaults = {'step_bound': 100.0, 'scale_diag': 1}

        def get_value(option_name):
            value = keyword_values.get(option_name)
            if value is None:
                value = optimizer_settings.get(option_name)
            if value is None:
                value = option_defaults[option_name]
            return value

        for option in option_defaults:
            optimizer_settings.update({option: get_value(option)})

        super(LevenbergMarquardt, self).__init__(patience=patience, optimizer_settings=optimizer_settings, **kwargs)

    def _get_worker_generator(self, *args):
        return lambda cl_environment: LevenbergMarquardtWorker(cl_environment, *args)


class LevenbergMarquardtWorker(AbstractParallelOptimizerWorker):

    def __init__(self, *args, **kwargs):
        super(LevenbergMarquardtWorker, self).__init__(*args, **kwargs)

        if self._model.get_nmr_inst_per_problem() < self._nmr_params:
            raise ValueError('The number of instances per problem must be greater than the number of parameters')

    def _create_buffers(self):
        all_buffers, parameters_buffer, return_code_buffer = super(LevenbergMarquardtWorker, self)._create_buffers()

        fjac_items = self._nmr_params * self._model.get_nmr_inst_per_problem() * self._starting_points.shape[0]
        fjac_buffer_size = fjac_items * self._starting_points.dtype.itemsize
        fjac_buffer = cl.Buffer(self._cl_context, cl.mem_flags.READ_WRITE, size=fjac_buffer_size)
        all_buffers.append(fjac_buffer)

        return all_buffers, parameters_buffer, return_code_buffer

    def _get_buffer_scalar_args(self):
        scalar_args = super(LevenbergMarquardtWorker, self)._get_buffer_scalar_args()
        scalar_args.insert(2, None)
        return scalar_args

    def _get_kernel_source(self):
        """Overwrite the default kernel source generation.

        We need to add a few extra buffers to the kernel and the lmmin call. This makes the optimizer
        more robust when using datasets with a high number of observations per problem instance.

        If we inline these buffers the CL compiler may throw a 'out of resources' error since it can fail to find
        a contiguous block of memory large enough for any of these buffers. We circumvent that now by creating them
        in global memory space.
        """
        nmr_params = self._nmr_params

        kernel_param_names = ['global mot_float_type* restrict params',
                              'global char* restrict return_codes']
        kernel_param_names.extend(self._data_struct_manager.get_kernel_arguments())
        kernel_param_names.append('global mot_float_type* restrict fjac_all')

        optimizer_call_args = 'x, (void*)&data'

        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += self._data_struct_manager.get_struct_definition()

        kernel_source += self._get_optimizer_cl_code()
        kernel_source += '''
            __kernel void minimize(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    ulong gid = get_global_id(0);

                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                    }

                    global mot_float_type* fjac = fjac_all + gid * ''' \
                         + str(self._nmr_params * self._model.get_nmr_inst_per_problem()) + ''';

                    mot_data_struct data = ''' + self._data_struct_manager.get_struct_init_string('gid') + ''';
                    return_codes[gid] = (char) ''' + self._get_optimizer_call_name() + '''(''' \
                         + optimizer_call_args + ''', fjac);

                    for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                        params[gid * ''' + str(nmr_params) + ''' + i] = x[i];
                    }
                }
        '''
        return kernel_source

    def _get_evaluate_function(self):
        """Get the CL code for the evaluation function. This is called from _get_optimizer_cl_code.

        Implementing optimizers can change this if desired.

        Returns:
            str: the evaluation function.
        """
        objective_func = self._model.get_objective_per_observation_function()
        param_modifier = self._model.get_pre_eval_parameter_modifier()

        kernel_source = ''
        kernel_source += objective_func.get_cl_code()
        kernel_source += param_modifier.get_cl_code()
        kernel_source += '''
            void evaluate(mot_float_type* x, void* data_void, mot_float_type* result){
                mot_data_struct* data = (mot_data_struct*)data_void;
                
                mot_float_type x_model[''' + str(self._model.get_nmr_estimable_parameters()) + '''];
                for(uint i = 0; i < ''' + str(self._model.get_nmr_estimable_parameters()) + '''; i++){
                    x_model[i] = x[i];
                }
                ''' + param_modifier.get_cl_function_name() + '''(data, x_model);
                
                for(uint i = 0; i < ''' + str(self._model.get_nmr_inst_per_problem()) + '''; i++){
                    // the model expects the L1 norm, while the LM method takes the L2 norm. Taking square root here.
                    result[i] = sqrt(fabs(''' + objective_func.get_cl_function_name() + '''(data, x_model, i)));
                }
            }
        '''
        return kernel_source

    def _get_optimization_function(self):
        params = {'NMR_PARAMS': self._nmr_params,
                  'PATIENCE': self._parent_optimizer.patience,
                  'NMR_INST_PER_PROBLEM': self._model.get_nmr_inst_per_problem(),
                  'USER_TOL_MULT': 30}

        optimizer_settings = self._optimizer_settings or {}
        option_defaults = {'step_bound': 100.0, 'scale_diag': 1, 'usertol_mult': 30}
        option_converters = {'scale_diag': lambda val: int(bool(val))}

        for option, default in option_defaults.items():
            v = optimizer_settings.get(option, default)
            if option in option_converters:
                v = option_converters[option](v)
            params.update({option.upper(): v})

        body = open(os.path.abspath(resource_filename('mot', 'data/opencl/lmmin.cl')), 'r').read()
        if params:
            body = body % params
        return body

    def _get_optimizer_call_name(self):
        return 'lmmin'



from mot.cl_data_type import SimpleCLDataType
from mot.cl_parameter import SimpleCLFunctionParameter
from mot.model_building.parameter_functions.numdiff_info import SimpleNumDiffInfo
from mot.model_building.parameter_functions.priors import UniformWithinBoundsPrior
from mot.model_building.parameter_functions.transformations import IdentityTransform

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CurrentObservationParam(SimpleCLFunctionParameter):

    def __init__(self, name='_observation'):
        """This parameter indicates that the model should inject the current observation value in the model.

        Sometimes during model linearization or other mathematical operations the current observation appears on
        both sides of the optimization equation. That is, it sometimes happens you want to use the current observation
        to model that same observation. This parameter is a signal to the model builder to inject the current
        observation.

        You can use this parameter by adding it to your model and then use the current name in your model equation.
        """
        super(CurrentObservationParam, self).__init__(SimpleCLDataType.from_string('mot_float_type'), name)


class InputDataParameter(SimpleCLFunctionParameter):

    def __init__(self, data_type, name, value):
        """These parameters signal are meant to be contain data loaded from the input data object.

        In contrast to free parameters which are being optimized (or fixed to values), these parameters are
        meant to be loaded from the input data. They can contain scalars, vectors or matrices with values
        to use for each problem instance and each data point.

        Args:
            data_type (mot.cl_data_type.SimpleCLDataType): the data type expected by this parameter
            name (str): The name of this parameter
            value (double or ndarray): The value used if no value is given in the input data.
        """
        super(InputDataParameter, self).__init__(data_type, name)
        self.value = value


class StaticMapParameter(InputDataParameter):

    def __init__(self, data_type, name, value):
        """This parameter is meant for static data that is different per problem.

        These parameters are in usage similar to fixed free parameters. They are defined as static data parameters to
        make clear that they are meant to provide additional observational data.

        Values for this parameter type are typically loaded from the input data. A default is provided in the case
        that there is no suitable value in the input data.

        Args:
            data_type (mot.cl_data_type.SimpleCLDataType): the data type expected by this parameter
            name (str): The name of this parameter
            value (double or ndarray): The value used if no value is given in the input data.
        """
        super(StaticMapParameter, self).__init__(data_type, name, value)
        self.value = value


class ProtocolParameter(InputDataParameter):

    def __init__(self, data_type, name):
        """A protocol data parameter indicates that similar named parameters may be linked together.

        When multiple functions in a composite model share a protocol parameter with the same name, the model
        builder will only load the data once and link the identical named parameters to the same value.

        The value for a protocol parameter is loaded from the input data.
        """
        super(ProtocolParameter, self).__init__(data_type, name, None)


class FreeParameter(SimpleCLFunctionParameter):

    def __init__(self, data_type, name, fixed, value, lower_bound, upper_bound,
                 parameter_transform=None, sampling_proposal_std=None,
                 sampling_prior=None, numdiff_info=None):
        """This are the kind of parameters that are generally meant to be optimized.

        These parameters may optionally be fixed to a value or list of values for all problems.

        Args:
            data_type (str or mot.cl_data_type.SimpleCLDataType): the data type expected by this parameter
            name (str): The name of this parameter
            fixed (boolean): If this parameter is fixed to the value given
            value (double or ndarray): A single value for all problems or a list of values for each problem.
            lower_bound (double): The lower bound of this parameter
            upper_bound (double): The upper bound of this parameter
            parameter_transform (AbstractTransformation): The parameter transformation function
            sampling_proposal_std (float): The proposal standard deviation, used in some MCMC sampling routines
            sampling_prior (ParameterPrior): The prior function for use in model sampling
            numdiff_info (mot.model_building.parameter_functions.numdiff_info.NumDiffInfo): the information
                for taking the numerical derivative with respect to this parameter.
        """
        super(FreeParameter, self).__init__(data_type, name)
        self._value = value
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._fixed = fixed

        self._parameter_transform = parameter_transform or IdentityTransform()
        self._sampling_proposal_std = sampling_proposal_std or 1
        self._sampling_prior = sampling_prior or UniformWithinBoundsPrior()
        self._numdiff_info = numdiff_info or SimpleNumDiffInfo()

    @property
    def value(self):
        return self._value

    @property
    def lower_bound(self):
        return self._lower_bound

    @property
    def upper_bound(self):
        return self._upper_bound

    @property
    def fixed(self):
        return self._fixed

    @property
    def parameter_transform(self):
        return self._parameter_transform

    @property
    def sampling_proposal_std(self):
        return self._sampling_proposal_std

    @property
    def sampling_prior(self):
        return self._sampling_prior

    @property
    def numdiff_info(self):
        return self._numdiff_info


class LibraryParameter(SimpleCLFunctionParameter):
    """Parameters of this type are used inside library functions. They are not meant to be used in Model functions.
    """



from mot.cl_routines.optimizing.multi_step_optimizer import MultiStepOptimizer
from mot.cl_routines.optimizing.random_restart import RandomRestart
from mot.cl_routines.optimizing.sbplex import SBPlex
from mot.cl_routines.sampling.amwg import AdaptiveMetropolisWithinGibbs
from .cl_routines.optimizing.levenberg_marquardt import LevenbergMarquardt
from .cl_routines.filters.gaussian import GaussianFilter
from .cl_routines.filters.mean import MeanFilter
from .cl_routines.filters.median import MedianFilter
from .cl_routines.optimizing.nmsimplex import NMSimplex
from .cl_routines.optimizing.powell import Powell
from .load_balance_strategies import EvenDistribution, RuntimeLoadBalancing, PreferGPU, PreferCPU, \
    PreferSpecificEnvironment

__author__ = 'Robbert Harms'
__date__ = "2015-07-06"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


optimizers = [LevenbergMarquardt, Powell, NMSimplex, SBPlex, MultiStepOptimizer, RandomRestart]
samplers = [AdaptiveMetropolisWithinGibbs]
filters = [GaussianFilter, MeanFilter, MedianFilter]
load_balance_strategies = [EvenDistribution, RuntimeLoadBalancing, PreferGPU, PreferCPU, PreferSpecificEnvironment]


def get_optimizer_by_name(name):
    """ Get the class by the given name.

    This does not instantiate the class, only returns a reference to it.

    Args:
        name: the name of the optimizer we want to return

    Returns:
        class: the class of the optimizer requested
    """
    return _get_item(name, optimizers, 'optimizers')


def get_sampler_by_name(name):
    """ Get the class by the given name.

    This does not instantiate the class, only returns a reference to it.

    Args:
        name: the name of the optimizer we want to return

    Returns:
        class: the class of the sampler requested
    """
    return _get_item(name, samplers, 'samplers')


def get_filter_by_name(name):
    """ Get the class by the given name.

    This does not instantiate the class, only returns a reference to it.

    Args:
        name: the name of the filter routine we want to return

    Returns:
        class: the class of the filter routine requested
    """
    return _get_item(name, filters, 'smoothers')


def get_load_balance_strategy_by_name(name):
    """ Get the class by the given name.

    This does not instantiate the class, only returns a reference to it.

    Args:
        name: the name of the load balance strategy we want to return

    Returns:
        class: the class of the load balance strategy requested
    """
    return _get_item(name, load_balance_strategies, 'load balancers')


def _get_item(name, item_list, factory_type):
    for item in item_list:
        if item.__name__ == name:
            return item
    raise ValueError('The item with the name {0} could not be found in the {1} factory.'.format(name, factory_type))



from mot.cl_routines.mapping.run_procedure import RunProcedure
from ...utils import KernelInputArray, KernelInputScalar, SimpleNamedCLFunction, KernelInputLocalMemory, \
    KernelInputAllocatedOutput
from ...cl_routines.base import CLRoutine


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class LogLikelihoodCalculator(CLRoutine):

    def calculate(self, model, parameters):
        """Calculate and return the log likelihood of the given model under the given parameters.

        This calculates log likelihoods for every problem in the model (typically after optimization),
        or a log likelihood for every sample of every model (typical after sampling). In the case of the first you
        can provide this function with a dictionary of parameters, or with an (d, p) array with d problems and
        p parameters. In the case of the second (after sampling), you must provide this function with a matrix of shape
        (d, p, n) with d problems, p parameters and n samples.

        Args:
            model (AbstractModel): The model to calculate the full log likelihood for.
            parameters (ndarray): The parameters to use in the evaluation of the model. This is either an (d, p) matrix
                or (d, p, n) matrix with d problems, p parameters and n samples.

        Returns:
            ndarray: per problem the log likelihood, or, per problem per sample the calculate log likelihood.
        """
        all_kernel_data = dict(model.get_kernel_data())
        all_kernel_data.update({
            'parameters': KernelInputArray(parameters),
        })

        shape = parameters.shape
        if len(shape) > 2:
            all_kernel_data.update({
                'log_likelihoods': KernelInputAllocatedOutput((shape[0], shape[2]), 'mot_float_type'),
                'nmr_params': KernelInputScalar(parameters.shape[1]),
                'nmr_samples': KernelInputScalar(parameters.shape[2]),
                'local_reduction_lls': KernelInputLocalMemory('double')
            })
        else:
            all_kernel_data.update({
                'log_likelihoods': KernelInputAllocatedOutput((shape[0],), 'mot_float_type'),
                'local_reduction_lls': KernelInputLocalMemory('double')
            })

        runner = RunProcedure(**self.get_cl_routine_kwargs())
        runner.run_procedure(self._get_wrapped_function(model, parameters), all_kernel_data, parameters.shape[0],
                             double_precision=model.double_precision, use_local_reduction=True)

        return all_kernel_data['log_likelihoods'].get_data()

    def _get_wrapped_function(self, model, parameters):
        ll_func = model.get_log_likelihood_per_observation_function()
        nmr_params = parameters.shape[1]

        func = ''
        func += ll_func.get_cl_code()
        func += '''
            void _fill_log_likelihood_tmp(mot_data_struct* data,
                                          mot_float_type* x,
                                          local double* log_likelihood_tmp){

                ulong observation_ind;
                ulong local_id = get_local_id(0);
                log_likelihood_tmp[local_id] = 0;
                uint workgroup_size = get_local_size(0);
                uint elements_for_workitem = ceil(''' + str(model.get_nmr_inst_per_problem()) + ''' 
                                                  / (mot_float_type)workgroup_size);
                
                if(workgroup_size * (elements_for_workitem - 1) + local_id 
                        >= ''' + str(model.get_nmr_inst_per_problem()) + '''){
                    elements_for_workitem -= 1;
                }
                
                for(uint i = 0; i < elements_for_workitem; i++){
                    observation_ind = i * workgroup_size + local_id;
                
                    log_likelihood_tmp[local_id] += ''' + ll_func.get_cl_function_name() + '''(
                        data, x, observation_ind);
                }

                barrier(CLK_LOCAL_MEM_FENCE);
            }

            double _sum_log_likelihood_tmp(local double* log_likelihood_tmp){
                double ll = 0;
                for(uint i = 0; i < get_local_size(0); i++){
                    ll += log_likelihood_tmp[i];
                }
                return ll;
            }
            
        '''

        if len(parameters.shape) > 2:
            func += '''
                void compute(mot_data_struct* data){
                    mot_float_type x[''' + str(nmr_params) + '''];
                    
                    for(uint sample_ind = 0; sample_ind < ''' + str(parameters.shape[2]) + '''; sample_ind++){
                        for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                            x[i] = data->parameters[i *''' + str(parameters.shape[2]) + ''' + sample_ind];
                        }

                        _fill_log_likelihood_tmp(data, x, data->local_reduction_lls);
                        if(get_local_id(0) == 0){
                            data->log_likelihoods[sample_ind] = _sum_log_likelihood_tmp(data->local_reduction_lls);
                        }
                    }
                }
            '''
        else:
            func += '''
                void compute(mot_data_struct* data){
                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = data->parameters[i];
                    }
                    
                    _fill_log_likelihood_tmp(data, x, data->local_reduction_lls);
                    if(get_local_id(0) == 0){
                        *(data->log_likelihoods) = _sum_log_likelihood_tmp(data->local_reduction_lls);
                    }
                }
            '''
        return SimpleNamedCLFunction(func, 'compute')



from ...utils import get_float_type_def, KernelInputDataManager
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker
import pyopencl as cl
import numpy as np

__author__ = 'Robbert Harms'
__date__ = '2017-08-31'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class RunProcedure(CLRoutine):

    def __init__(self, **kwargs):
        """This class can run any arbitrary given CL procedure on the given set of data."""
        super(RunProcedure, self).__init__(**kwargs)

    def run_procedure(self, cl_function, kernel_data, nmr_instances, double_precision=False,
                      use_local_reduction=False):
        """Run the given function/procedure on the given set of data.

        This class will wrap the given CL function in a kernel call and execute that that for every data instance using
        the provided kernel data. This class will respect the read write setting of the kernel data elements such that
        output can be written back to the according kernel data elements.

        Args:
            cl_function (mot.utils.NamedCLFunction): the function to run on the datasets
            kernel_data (dict[str: mot.utils.KernelInputData]): the data to use as input to the function
                all the data will be wrapped in a single ``mot_data_struct``.
            nmr_instances (int): the number of parallel threads to run (used as ``global_size``)
            double_precision (boolean): if we want to run in double precision. Defaults to True.
            use_local_reduction (boolean): set this to True if you want to use local memory reduction in
                 your CL procedure. If this is set to True we will multiply the global size (given by the nmr_instances)
                 by the work group sizes.
        """
        workers = self._create_workers(lambda cl_environment: _ProcedureWorker(
            cl_environment, self.get_compile_flags_list(True),
            cl_function, kernel_data, double_precision, use_local_reduction))
        self.load_balancer.process(workers, nmr_instances)


class _ProcedureWorker(Worker):

    def __init__(self, cl_environment, compile_flags, named_cl_function, kernel_data, double_precision,
                 use_local_reduction):
        super(_ProcedureWorker, self).__init__(cl_environment)
        self._cl_func = named_cl_function.get_cl_code()
        self._cl_func_name = named_cl_function.get_cl_function_name()
        self._kernel_data = kernel_data
        self._double_precision = double_precision
        self._use_local_reduction = use_local_reduction

        mot_float_dtype = np.float32
        if double_precision:
            mot_float_dtype = np.float64

        self._data_struct_manager = KernelInputDataManager(self._kernel_data, mot_float_dtype)
        self._kernel = self._build_kernel(self._get_kernel_source(), compile_flags)
        self._workgroup_size = self._kernel.run_procedure.get_work_group_info(
            cl.kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE,
            self._cl_environment.device)
        if not self._use_local_reduction:
            self._workgroup_size = 1

        self._kernel_input = self._get_kernel_input()

    def _get_kernel_input(self):
        return self._data_struct_manager.get_kernel_inputs(self._cl_context, self._workgroup_size)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        func = self._kernel.run_procedure
        func.set_scalar_arg_dtypes(self._data_struct_manager.get_scalar_arg_dtypes())
        func(self._cl_queue,
             (int(nmr_problems * self._workgroup_size), ),
             (int(self._workgroup_size),),
             *self._kernel_input,
             global_offset=(int(range_start * self._workgroup_size),))

        for ind, name in self._data_struct_manager.get_items_to_write_out():
            self._enqueue_readout(self._kernel_input[ind], self._kernel_data[name].get_data(), range_start, range_end)

    def _get_kernel_source(self):
        kernel_param_names = self._data_struct_manager.get_kernel_arguments()

        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += self._data_struct_manager.get_struct_definition()
        kernel_source += self._cl_func
        kernel_source += '''
            __kernel void run_procedure(
                    ''' + ",\n".join(kernel_param_names) + '''){

                ulong gid = ''' + ('(ulong)(get_global_id(0) / get_local_size(0));'
                                   if self._use_local_reduction else 'get_global_id(0)') + ''';

                mot_data_struct data = ''' + self._data_struct_manager.get_struct_init_string('gid') + ''';
                ''' + self._cl_func_name + '''(&data);
            }
        '''
        return kernel_source



from mot.cl_routines.mapping.run_procedure import RunProcedure
from ...utils import SimpleNamedCLFunction, KernelInputArray, KernelInputAllocatedOutput
from ...cl_routines.base import CLRoutine
import numpy as np

__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CalculateDependentParameters(CLRoutine):

    def __init__(self, double_precision=False, **kwargs):
        """CL code for calculating the dependent parameters.

        Some of the models may contain parameter dependencies. We would like to return the maps for these parameters
        as well as all the other maps. Since the dependencies are specified in CL, we have to recourse to CL to
        calculate these maps.

        Args:
            double_precision (boolean): if we will use the double (True) or single floating (False) type
                for the calculations
        """
        super(CalculateDependentParameters, self).__init__(**kwargs)
        self._double_precision = double_precision

    def calculate(self, kernel_data, estimated_parameters_list, parameters_listing, dependent_parameter_names):
        """Calculate the dependent parameters

        This uses the calculated parameters in the results dictionary to run the parameters_listing in CL to obtain
        the maps for the dependent parameters.

        Args:
            kernel_data (dict[str: mot.utils.KernelInputData]): the list of additional data to load
            estimated_parameters_list (list of ndarray): The list with the one-dimensional
                ndarray of estimated parameters
            parameters_listing (str): The parameters listing in CL
            dependent_parameter_names (list of list of str): Per parameter we would like to obtain the CL name and the
                result map name. For example: (('Wball_w', 'Wball.w'),)
        Returns:
            dict: A dictionary with the calculated maps for the dependent parameters.
        """
        cl_named_func = self._get_wrapped_function(estimated_parameters_list, parameters_listing,
                                                   dependent_parameter_names)

        all_kernel_data = dict(kernel_data)
        all_kernel_data['x'] = KernelInputArray(np.dstack(estimated_parameters_list)[0, ...], ctype='mot_float_type')
        all_kernel_data['_results'] = KernelInputAllocatedOutput(
            (estimated_parameters_list[0].shape[0], len(dependent_parameter_names)), 'mot_float_type')

        runner = RunProcedure(**self.get_cl_routine_kwargs())
        runner.run_procedure(cl_named_func, all_kernel_data, estimated_parameters_list[0].shape[0],
                             double_precision=self._double_precision)

        return all_kernel_data['_results'].get_data()

    def _get_wrapped_function(self, estimated_parameters_list, parameters_listing, dependent_parameter_names):
        parameter_write_out = ''
        for i, p in enumerate([el[0] for el in dependent_parameter_names]):
            parameter_write_out += 'data->_results[' + str(i) + '] = ' + p + ";\n"

        func = '''
            void transform(mot_data_struct* data){
                mot_float_type x[''' + str(len(estimated_parameters_list)) + '''];

                for(uint i = 0; i < ''' + str(len(estimated_parameters_list)) + '''; i++){
                    x[i] = data->x[i];
                }
                ''' + parameters_listing + '''
                ''' + parameter_write_out + '''
            }
        '''
        return SimpleNamedCLFunction(func, 'transform')



__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class InputData(object):
    """A simple container for the input data for optimization/sampling models."""

    def get_input_data(self, parameter_name):
        """Get the input data for the given parameter.

        Args:
             parameter_name (str): the name of the parameter for which we want to get input data

        Returns:
            float, ndarray or None: either a scalar, a vector or a matrix with values for the given parameter.
                None should be returned if no suitable value can be found.
        """
        raise NotImplementedError()

    @property
    def nmr_problems(self):
        """Get the number of problems present in this input data.

        Returns:
            int: the number of problem instances
        """
        raise NotImplementedError()

    @property
    def nmr_observations(self):
        """Get the number of observations/data points per problem.

        The minimum is one observation sper problem.

        Returns:
            int: the number of instances per problem (aka data points)
        """
        raise NotImplementedError()

    @property
    def observations(self):
        """Return the observations stored in this input data container.

        Returns:
            ndarray: The list of observed instances per problem. Should be a 2d matrix of type float with as
                columns the observations and as rows the problems.
        """
        raise NotImplementedError()

    @property
    def noise_std(self):
        """The noise standard deviation we will use during model evaluation.

        During optimization or sampling the model will be evaluated against the observations using a
        likelihood function. Most of these likelihood functions need a standard deviation representing the noise
        in the data.

        Returns:
            number of ndarray: either a scalar or a 2d matrix with one value per problem instance.
        """
        raise NotImplementedError()



from mot.cl_routines.mapping.run_procedure import RunProcedure
from ...utils import SimpleNamedCLFunction, KernelInputArray, KernelInputScalar, KernelInputAllocatedOutput
from ...cl_routines.base import CLRoutine
import numpy as np

__author__ = 'Robbert Harms'
__date__ = '2017-09-11'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class GaussianFit(CLRoutine):

    def calculate(self, samples, ddof=1, return_variance=False):
        """Calculate the mean and standard deviation.

        This expects a two dimensional array as input and calculates the mean and std similar as doing::

            np.mean(samples, axis=1)
            np.std(samples, axis=1, ddof=ddof)

        That is, by default it will calculate the sample variance and not the population variance.

        Args:
            samples (ndarray): the samples from which we want to calculate the mean and std.
            ddof (ddof): the difference degree of freedom
            return_variance (boolean): if we want to return the variance instead of the standard deviation

        Returns:
            tuple: mean and deviation arrays
        """
        all_kernel_data = {
            'samples': KernelInputArray(samples, ctype='mot_float_type'),
            'means': KernelInputAllocatedOutput(samples.shape[0], 'mot_float_type'),
            'deviations': KernelInputAllocatedOutput(samples.shape[0], 'mot_float_type'),
            'nmr_samples': KernelInputScalar(samples.shape[1]),
            'ddof': KernelInputScalar(ddof)
        }

        runner = RunProcedure(**self.get_cl_routine_kwargs())
        runner.run_procedure(self._get_wrapped_function(return_variance), all_kernel_data, samples.shape[0],
                             double_precision=(samples.dtype == np.float64))

        return all_kernel_data['means'].get_data(), all_kernel_data['deviations'].get_data()

    def _get_wrapped_function(self, return_variance):
        func = '''
            void compute(mot_data_struct* data){
                double variance = 0;
                double mean = 0;
                double delta;
                double value;
                
                for(uint i = 0; i < data->nmr_samples; i++){
                    value = data->samples[i];
                    delta = value - mean;
                    mean += delta / (i + 1);
                    variance += delta * (value - mean);
                }
                variance /= (data->nmr_samples - data->ddof);
                
                *(data->means) = mean;
                *(data->deviations) = ''' + ('variance' if return_variance else 'sqrt(variance)') + ''';
            }            
        '''
        return SimpleNamedCLFunction(func, 'compute')



from mot.cl_parameter import SimpleCLFunctionParameter
from mot.model_building.parameters import FreeParameter
from mot.model_building.model_functions import SimpleModelCLFunction
from mot.model_building.parameter_functions.transformations import CosSqrClampTransform

__author__ = 'Robbert Harms'
__date__ = "2014-08-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SignalNoiseModel(SimpleModelCLFunction):
    """Signal noise models can add noise to the signal resulting from the model.

    They require the signal resulting from the model and zero or more parameters and they return a new signal
    with noise added. This should have a model signature like:

    .. code-block:: c

        double fname(double signal, <noise model parameters ...>);

    For example, if the noise model has only one parameter 'sigma' the function should look like:

    .. code-block:: c

        double fname(double signal, double sigma);

    The CL function should return a single double that represents the signal with the signal noise
        added to it.
    """


class JohnsonNoise(SignalNoiseModel):

    def __init__(self):
        """Johnson noise adds noise to the signal using the formula:

        .. code-block:: c

            sqrt(signal^2 + eta^2)

        """
        super(JohnsonNoise, self).__init__(
            'double', 'JohnsonNoise', 'JohnsonNoise',
            [SimpleCLFunctionParameter('double', 'signal'),
             FreeParameter('mot_float_type', 'eta', False, 0.1, 0, 100, parameter_transform=CosSqrClampTransform())],
            'return hypot(signal, (double)eta);')



import six

__author__ = 'Robbert Harms'
__date__ = "2015-03-21"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CLDataType(object):
    """Interface for CL data type containers.

    Basically this encapsulates the type and its qualifiers that define a variable in CL.
    """

    def get_declaration(self):
        """Get the complete CL declaration for this datatype.

        Returns:
            str: the declaration for this data type.
        """
        raise NotImplementedError()

    @property
    def declaration_type(self):
        """Get the type of this parameter in CL language

        This only returns the parameter type (like ``double`` or ``int*`` or ``float4*`` ...). It does not include other
        qualifiers.

        Returns:
            str: The name of this data type
        """
        raise NotImplementedError()

    @property
    def raw_data_type(self):
        """Get the raw data type without the vector and pointer additions.

        For example, if the data type is float4*, we will only return float here.

        Returns:
            str: the raw CL data type
        """
        raise NotImplementedError()

    @property
    def ctype(self):
        """Get the ctype of this data type.

        Returns:
            str: the full ctype of this data type
        """
        raise NotImplementedError()

    @property
    def is_vector_type(self):
        """Check if this data type is a vector type (like for example double4, float2, int8, etc.).

        Returns:
            boolean: True if it is a vector type, false otherwise
        """
        raise NotImplementedError()

    @property
    def is_pointer_type(self):
        """Check if this parameter is a pointer type (appended by a ``*``)

        Returns:
            boolean: True if it is a pointer type, false otherwise
        """
        raise NotImplementedError()

    @property
    def vector_length(self):
        """Get the length of this vector, returns None if not a vector type.

        Returns:
            int: the length of the vector type (for example, if the data type is float4, this returns 4).
        """
        raise NotImplementedError()


class SimpleCLDataType(CLDataType):

    def __init__(self, raw_data_type, is_pointer_type=False, vector_length=None,
                 address_space_qualifier=None, pre_asterisk_qualifiers=None,
                 post_asterisk_qualifiers=None):
        """Create a new CL data type container.

        The CL type can either be a CL native type (``half``, ``double``, ``int``, ...) or the
        special ``mot_float_type`` type.

        Args:
            raw_data_type (str): the specific data type without the vector number and asterisks
            is_pointer_type (boolean): If this parameter is a pointer type (appended by a ``*``)
            vector_length (int or None): If None this data type is not a CL vector type.
                If it is an integer it is the vector length of this data type (2, 3, 4, ...)
            address_space_qualifier (str or None): the address space qualifier or None if not used. One of:
                {``__local``, ``local``, ``__global``, ``global``,
                ``__constant``, ``constant``, ``__private``, ``private``} or None.
            pre_asterisk_qualifiers (list of str or None): the type qualifiers to use before the (optional) asterisk.
                One or more of {const, volatile}
            post_asterisk_qualifiers (list of str or None): the type qualifiers to use after the (optional) asterisk.
                One or more of {const, restrict}
        """
        self._raw_data_type = str(raw_data_type)
        self._is_pointer_type = is_pointer_type
        self._vector_length = vector_length

        if self.vector_length:
            self._vector_length = int(self.vector_length)

        self.address_space_qualifier = address_space_qualifier

        self.pre_asterisk_qualifiers = pre_asterisk_qualifiers
        if isinstance(self.pre_asterisk_qualifiers, six.string_types):
            self.pre_asterisk_qualifiers = [self.pre_asterisk_qualifiers]

        self.post_asterisk_qualifiers = post_asterisk_qualifiers
        if isinstance(self.post_asterisk_qualifiers, six.string_types):
            self.post_asterisk_qualifiers = [post_asterisk_qualifiers]

    @classmethod
    def from_string(cls, parameter_declaration):
        """Parse the parameter declaration into a CLDataType

        Args:
            parameter_declaration (str): the CL parameter declaration. Example: ``global const float4*`` const

        Returns:
            mot.cl_data_type.SimpleCLDataType: the CL data type for this parameter declaration
        """
        from mot.parsers.cl.CLDataTypeParser import parse
        return parse(parameter_declaration)

    def get_declaration(self):
        declaration = ''
        if self.address_space_qualifier:
            declaration += str(self.address_space_qualifier) + ' '
        if self.pre_asterisk_qualifiers:
            declaration += str(' '.join(self.pre_asterisk_qualifiers)) + ' '
        declaration += str(self.declaration_type)
        if self.post_asterisk_qualifiers:
            declaration += ' ' + str(' '.join(self.post_asterisk_qualifiers)) + ' '
        return declaration

    @property
    def declaration_type(self):
        s = self.raw_data_type

        if self.vector_length is not None:
            s += str(self.vector_length)

        if self.is_pointer_type:
            s += '*'

        return str(s)

    @property
    def is_vector_type(self):
        return self.vector_length is not None

    @property
    def is_pointer_type(self):
        return self._is_pointer_type

    @property
    def vector_length(self):
        return self._vector_length

    @property
    def raw_data_type(self):
        return self._raw_data_type

    @property
    def ctype(self):
        if self.is_vector_type:
            return self._raw_data_type + str(self.vector_length)
        return self._raw_data_type

    def __str__(self):
        return self.get_declaration()



import numpy as np
from mot.cl_routines.optimizing.base import AbstractOptimizer, SimpleOptimizationResult

__author__ = 'Robbert Harms'
__date__ = "2016-11-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class StartingPointGenerator(object):

    def next(self, model, previous_results):
        """Returns a next starting point given the model and the previous results.

        This method can return None which means that no next starting point is available.

        Args:
            model: the model for which we are generating points
            previous_results (ndarray): the previous results, an (d, p) array with for every d problems and n parameters
                the (semi-)optimum value

        Returns:
            ndarray: array of same type and shape as the input but with the new starting points
        """
        return NotImplementedError


class PointsFromGrid(StartingPointGenerator):

    def __init__(self, grid):
        """Uses the given points as a starting point for all problem instances per run.

        This accepts a grid with per row a starting point that we will use for all the problem instances for
        that optimization run. The number of rows determines the number of iterations.

        Args:
            grid (ndarray): the grid with one starting point per iteration.
        """
        self.grid = grid
        self._iteration_counter = 0

    def next(self, model, previous_results):
        if self._iteration_counter == self.grid.shape[0]:
            return None

        new_points = np.zeros_like(previous_results)

        for param_ind in range(new_points.shape[1]):
            new_points[:, param_ind] = np.ones(model.get_nmr_problems()) * self.grid[self._iteration_counter, param_ind]

        self._iteration_counter += 1
        return new_points


class UsePrevious(StartingPointGenerator):

    def __init__(self, number_of_runs=3):
        """A strategy that uses the previous results without alterations for the next optimization run.

        Args:
            number_of_runs (int): the number of times we optimize using this strategy
        """
        self.number_of_runs = number_of_runs
        self._iteration_counter = 0

    def next(self, model, previous_results):
        if self._iteration_counter == self.number_of_runs:
            return None
        self._iteration_counter += 1
        return previous_results


class RandomStartingPoint(StartingPointGenerator):

    def __init__(self, number_of_runs=3):
        """A strategy that generates uniformly random starting points for every parameter.

        Per run this generates for each parameter a uniformly distributed random number between the lower
        and upper bounds and uses that single random value for all problem instances.

        Hence, this does not generate a unique random point per problem instance, but uses a single random
        point for all problem instances per iteration.

        Args:
            number_of_runs (int): the number of times we optimize using this strategy
        """
        self.number_of_runs = number_of_runs
        self._iteration_counter = 0

    def next(self, model, previous_results):
        if self._iteration_counter == self.number_of_runs:
            return None
        self._iteration_counter += 1

        lower_bounds = model.get_lower_bounds()
        upper_bounds = model.get_upper_bounds()

        new_points = np.zeros_like(previous_results)
        for param_ind in range(new_points.shape[1]):
            new_points[:, param_ind] = np.ones(model.get_nmr_problems()) * \
                                    np.random.uniform(np.min(lower_bounds[param_ind]),
                                                      np.max(upper_bounds[param_ind]))
        return new_points


class GaussianPerturbation(StartingPointGenerator):

    def __init__(self, number_of_runs=3):
        """A strategy that perturbates the previous results using a Normal distribution

        Per run this generates for each parameter and for each problem instance a new starting position using
        the previous parameter as a mean and for standard deviation it uses the standard deviation over the problem
        instances.

        Hence, this generates a unique random point per problem instance, in contrast to some of the other strategies.

        Args:
            number_of_runs (int): the number of times we optimize using this strategy
        """
        self.number_of_runs = number_of_runs
        self._iteration_counter = 0

    def next(self, model, previous_results):
        if self._iteration_counter == self.number_of_runs:
            return None
        self._iteration_counter += 1

        lower_bounds = model.get_lower_bounds()
        upper_bounds = model.get_upper_bounds()

        new_points = np.zeros_like(previous_results)
        for param_ind in range(new_points.shape[1]):
            std = np.std(previous_results[:, param_ind])

            for problem_ind in range(model.get_nmr_problems()):
                new_points[problem_ind, param_ind] = np.clip(
                    np.random.normal(previous_results[problem_ind, param_ind], std),
                    lower_bounds[param_ind], upper_bounds[param_ind])
        return new_points


class RandomRestart(AbstractOptimizer):

    def __init__(self, optimizer, starting_point_generator, **kwargs):
        """A meta optimization routine that allows multiple random restarts.

        This meta optimizer runs the given optimizer multiple times using different starting positions for each run.
        The starting positions are obtained using a starting point generator which returns new starting points
        given the initial starting points.

        Please note that the initial starting point is always optimized first as a baseline reference.

        The returned results contain per problem instance the parameter that resulted in the lowest function value.

        Args:
            optimizer (AbstractOptimizer): the optimization routines to run one after another.
            starting_point_generator (StartingPointGenerator): the randomizer instance we use
                to randomize the starting point
        """
        super(RandomRestart, self).__init__(**kwargs)
        self._optimizer = optimizer
        self._starting_point_generator = starting_point_generator

    def minimize(self, model, starting_positions):
        opt_output = self._optimizer.minimize(model, starting_positions)
        objective_value = opt_output.get_objective_values()
        results = opt_output.get_optimization_result()
        return_codes = opt_output.get_return_codes()

        starting_points = self._starting_point_generator.next(model, results)
        while starting_points is not None:
            new_opt_output = self._optimizer.minimize(model, starting_points)
            new_results = new_opt_output.get_optimization_result()
            new_objective_value = new_opt_output.get_objective_values()
            return_codes = new_opt_output.get_return_codes()

            results, objective_value = self._get_best_results(results, new_results, objective_value, new_objective_value)

            starting_points = self._starting_point_generator.next(model, results)

        return SimpleOptimizationResult(model, results, return_codes)

    def _get_best_results(self, previous_results, new_results, previous_objective_value, new_objective_value):
        result_choice = np.argmin([previous_objective_value, new_objective_value], axis=0)

        results = np.zeros_like(previous_results)

        for param_ind in range(previous_results.shape[1]):
            choices = np.array([previous_results[:, param_ind], new_results[:, param_ind]])
            results[:, param_ind] = choices[(result_choice, range(result_choice.shape[0]))]

        best_objective_value = np.array([previous_objective_value,
                                         new_objective_value])[(result_choice, range(result_choice.shape[0]))]

        return results, best_objective_value



"""The interfaces needed for models.

Since a lot of information about a model is needed to be able to optimize or sample it, we encapsulate all that
information in an interface. Only objects that successful implement the interfaces in this module can be optimized or
sampled using one of the optimization or sampling routines in MOT.

These interfaces expose data and modeling code. The data is represented as numpy arrays and the CL code as strings.
"""

__author__ = 'Robbert Harms'
__date__ = "2014-03-14"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class OptimizeModelInterface(object):

    @property
    def name(self):
        """Get the name of this model. This should be overwritten by the implementing model.

        Returns:
            str: A string with the name of this model.
        """
        raise NotImplementedError()

    @property
    def double_precision(self):
        """Flag to signal if we should use the double float type during calculations.

        By default we ask the cl routines to use the single precision float type, you can overwrite this with
        your own flags.

        Returns:
            boolean: if we would like to use double precision floating point during the calculations
        """
        raise NotImplementedError()

    def get_kernel_data(self):
        """Return a dictionary of input data objects we need to load into the kernel.

        Returns:
            dict[str: mot.utils.KernelInputData]: the list of data objects we need to load in the kernel
        """
        raise NotImplementedError()

    def get_nmr_problems(self):
        """Get the number of problems we need to analyze.

        Returns:
            int: A single integer specifying the number of problem instances
        """
        raise NotImplementedError()

    def get_nmr_inst_per_problem(self):
        """Get the number of instances/data points per problem.

        The minimum is one instance per problem.

        This number represents the number of data points

        Returns:
            int: the number of instances per problem.
        """
        raise NotImplementedError()

    def get_nmr_estimable_parameters(self):
        """Get the number of estimable parameters.

        Returns:
            int: the number of estimable parameters
        """
        raise NotImplementedError()

    def get_pre_eval_parameter_modifier(self):
        """Return code that needs to be run prior to model evaluation or objective function calculation.

        This is meant to contain possible parameter transformations that need to be executed only once for a
        given set of parameters. Having this in a separate function gives a speed gain.

        Model optimization routines need to be aware that they need to call this function prior to calling any of:

        * :meth:`~get_model_eval_function`
        * :meth:`~get_objective_per_observation_function`


        Returns:
            mot.utils.NamedCLFunction: a named CL function with the following signature:

                .. code-block:: c

                    void <func_name>(mot_data_struct* data, mot_float_type* x);

                Changes may happen in place in the ``x`` parameter.
        """
        raise NotImplementedError()

    def get_model_eval_function(self):
        """Get the evaluation function that evaluates the model at the given parameters.

        The returned function should not do any error calculations,
        it should merely return the result of evaluating the model for the given parameters.

        Returns:
            mot.utils.NamedCLFunction: a named CL function with the following signature:

                .. code-block:: c

                    double <func_name>(mot_data_struct* data, const mot_float_type* const x, uint observation_index);
        """
        raise NotImplementedError()

    def get_objective_per_observation_function(self):
        """Get the objective function that returns the objective value at a measurement instance.

        This should return the objective values (of each instance point) as such that when linearly summed, we have the
        complete objective value.

        Returns:
            mot.utils.NamedCLFunction: A CL function with signature:

                .. code-block:: c

                    double <func_name>(mot_data_struct* data, const mot_float_type* const x, uint observation_index);
        """
        raise NotImplementedError()

    def get_lower_bounds(self):
        """Get for each estimable parameter the lower bounds.

        Returns:
            list: For every estimable parameter a scalar or vector with the the lower bound(s) for that parameter.
                For infinity use np.inf.
        """
        raise NotImplementedError()

    def get_upper_bounds(self):
        """Get for each estimable parameter the upper bounds.

        Returns:
            list: For every estimable parameter a scalar or vector with the the upper bound(s) for that parameter.
                For infinity use np.inf.
        """
        raise NotImplementedError()

    def finalize_optimized_parameters(self, parameters):
        """Called by the optimization routine after optimization to finalize the optimized parameters.

        This can be used to do some automatic post-processing similar to for example what the
        ``get_pre_eval_parameter_modifier`` can do within the kernel.

        Args:
            parameters (ndarray): the set of parameters after optimization. This method may change this array
                in place.

        Returns:
            ndarray: the updated parameters. While changes may be done in place, one must return the parameters
                one would like to use.
        """
        raise NotImplementedError()


class SampleModelInterface(OptimizeModelInterface):
    """Extends the OptimizeModelInterface with information for sampling purposes."""

    def get_log_likelihood_per_observation_function(self):
        """Get the (complete) CL Log Likelihood function that evaluates the given instance under a noise model.

        This should return the LL's such that when linearly summed they yield the total log likelihood of the model.

        Returns:
            mot.utils.NamedCLFunction: A function of the kind:
                .. code-block:: c

                    double <fname>(mot_data_struct* data,
                                   const mot_float_type* const x,
                                   uint observation_index);
        """
        raise NotImplementedError()

    def get_log_prior_function(self, address_space_parameter_vector='private'):
        """Get the prior function that returns the prior information about the given parameters.

        The prior function must be in log space.

        Args:
            address_space_parameter_vector (str): the address space to use for the parameter vector
                by default this is set to ``private``.

        Returns:
            mot.utils.NamedCLFunction: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(
                        mot_data_struct* data,
                        <address_space_parameter_vector> const mot_float_type* const x
                    );

            Which is called by the sampling routine to calculate the prior probability.
        """
        raise NotImplementedError()

    def get_finalize_proposal_function(self, address_space_parameter_vector='private'):
        """Get a CL function that is called for every new proposal, in order to finalize it.

        This allows the model to change a proposal before computing the prior or likelihood probabilities.

        As an example, suppose you are sampling a polar coordinate :math:`\theta` defined on
        :math:`[0, 2\pi]` with a random walk Metropolis proposal distribution. This distribution might propose positions
        outside of the range of :math:`\theta`. Of course the model function could deal with that by taking the modulus
        of the input, but then you have to post-process the chain with the same transformation. Instead, this function
        allows changing the proposal before it is put into the model and before it is stored.

        Returns:
            mot.utils.NamedCLFunction: A function with the signature:
                .. code-block:: c

                    mot_float_type <func_name>(
                        mot_data_struct* data,
                        <address_space_parameter_vector> mot_float_type* x
                    );

            Which is called by the sampling routine to finalize the proposal.
        """
        raise NotImplementedError()


class NumericalDerivativeInterface(OptimizeModelInterface):
    """Extends the model with information necessary for calculating numerical derivatives of the objective function.

    For calculating derivatives (gradients / Hessians) numerically, we need a likelihood function and some additional
    information, like the step size for each parameter, a method for checking boundary conditions and possible parameter
    transformations for circular parameters. All these extra elements are represented in this interface.
    """

    def numdiff_get_max_step(self):
        """Get for each estimable parameter the maximum step size to use for calculating numerical derivatives.

        The derivative calculation method typically uses an adaptive step size to determine the step with the best
        trade-off between numerical errors and localization of the derivative. This method must return the
        initial and largest step size to use.

        Returns:
            list[float]: per parameter a single float with the step size for that parameter
        """
        raise NotImplementedError()

    def numdiff_get_scaling_factors(self):
        """Get for each estimable parameter a scaling factor that is to be used for scaling this parameter to unitary.

        Since numerical differentiation is sensitive to differences in step sizes, it is better to rescale
        the parameters to a unitary range instead of changing the step sizes for the parameters.

        This should return numbers such that when the parameter is multiplied with this value, the magnitude of the
        parameter is about one.

        Returns:
            list[float]: per estimable parameter a single float with the parameter scaling to use for that parameter.
                Use 1 as identity.
        """
        raise NotImplementedError()

    def numdiff_use_bounds(self):
        """Check for each parameter if we should be using the bounds for that parameter when taking the derivative.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_use_upper_bounds(self):
        """Check for each parameter if we should be using the upper bounds when taking the derivative.

        This is only used if use_bounds is True for a parameter.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the upper bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_use_lower_bounds(self):
        """Check for each parameter if we should be using the lower bounds when taking the derivative.

        This is only used if use_bounds is True for a parameter.

        Returns:
            list[bool]: per parameter a boolean to identify if we should use the lower bounds for that parameter.
        """
        raise NotImplementedError()

    def numdiff_parameter_transformation(self):
        """A transformation that can prepare the parameter +/- the proposed step for evaluation in the model function.

        Some parameters require for example a modulus operation before the proposed step can be used in the model
        This is the place to define them. Please note that this function does not need to incorporate boundary checks
        as those are handled already by the numerical differentiation routine.

        Returns:
            mot.utils.NamedCLFunction: A function with the signature:
                .. code-block:: c

                    void <func_name>(mot_data_struct* data, mot_float_type* params);

                Where the data is the kernel data struct and params is the vector with the suggested parameters and
                which can be modified in place. Note that this is called two times, one with the parameters plus
                the step and one time without.
        """
        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals

from grako.buffering import Buffer
from grako.parsing import graken, Parser


KEYWORDS = {}


class CLDataTypeBuffer(Buffer):
    def __init__(
        self,
        text,
        whitespace=None,
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        namechars='',
        **kwargs
    ):
        super(CLDataTypeBuffer, self).__init__(
            text,
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            namechars=namechars,
            **kwargs
        )


class CLDataTypeParser(Parser):
    def __init__(
        self,
        whitespace=None,
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        left_recursion=False,
        parseinfo=True,
        keywords=None,
        namechars='',
        buffer_class=CLDataTypeBuffer,
        **kwargs
    ):
        if keywords is None:
            keywords = KEYWORDS
        super(CLDataTypeParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            parseinfo=parseinfo,
            keywords=keywords,
            namechars=namechars,
            buffer_class=buffer_class,
            **kwargs
        )

    @graken()
    def _result_(self):
        self._expr_()

    @graken()
    def _expr_(self):
        with self._optional():
            self._address_space_qualifier_()

        def block0():
            self._pre_asterisk_qualifiers_()
        self._closure(block0)
        self._data_type_()
        with self._optional():
            self._is_pointer_()

            def block1():
                self._post_asterisk_qualifiers_()
            self._closure(block1)

    @graken()
    def _data_type_(self):
        with self._choice():
            with self._option():
                self._vector_data_type_()
            with self._option():
                self._user_data_type_()
            with self._option():
                self._scalar_data_type_()
            self._error('no available options')

    @graken()
    def _is_pointer_(self):
        self._token('*')

    @graken()
    def _scalar_data_type_(self):
        with self._choice():
            with self._option():
                self._token('mot_float_type')
            with self._option():
                self._token('bool')
            with self._option():
                self._token('char')
            with self._option():
                self._token('cl_char')
            with self._option():
                self._token('unsigned char')
            with self._option():
                self._token('uchar')
            with self._option():
                self._token('cl_uchar')
            with self._option():
                self._token('short')
            with self._option():
                self._token('cl_short')
            with self._option():
                self._token('unsigned short')
            with self._option():
                self._token('ushort')
            with self._option():
                self._token('int')
            with self._option():
                self._token('unsigned int')
            with self._option():
                self._token('uint')
            with self._option():
                self._token('long')
            with self._option():
                self._token('unsigned long')
            with self._option():
                self._token('ulong')
            with self._option():
                self._token('float')
            with self._option():
                self._token('half')
            with self._option():
                self._token('size_t')
            with self._option():
                self._token('ptrdiff_t')
            with self._option():
                self._token('intptr_t')
            with self._option():
                self._token('uintptr_t')
            with self._option():
                self._token('void')
            with self._option():
                self._token('double')
            with self._option():
                self._token('half')
            self._error('expecting one of: bool char cl_char cl_short cl_uchar double float half int intptr_t long mot_float_type ptrdiff_t short size_t uchar uint uintptr_t ulong unsigned char unsigned int unsigned long unsigned short ushort void')

    @graken()
    def _vector_data_type_(self):
        self._pattern(r'(char|uchar|short|ushort|int|uint|long|ulong|float|double|half|mot_float_type)\d+')

    @graken()
    def _user_data_type_(self):
        self._pattern(r'\w+')

    @graken()
    def _address_space_qualifier_(self):
        with self._choice():
            with self._option():
                self._token('__local')
            with self._option():
                self._token('local')
            with self._option():
                self._token('__global')
            with self._option():
                self._token('global')
            with self._option():
                self._token('__constant')
            with self._option():
                self._token('constant')
            with self._option():
                self._token('__private')
            with self._option():
                self._token('private')
            self._error('expecting one of: __constant __global __local __private constant global local private')

    @graken()
    def _pre_asterisk_qualifiers_(self):
        with self._choice():
            with self._option():
                self._token('const')
            with self._option():
                self._token('volatile')
            self._error('expecting one of: const volatile')

    @graken()
    def _post_asterisk_qualifiers_(self):
        with self._choice():
            with self._option():
                self._token('const')
            with self._option():
                self._token('restrict')
            self._error('expecting one of: const restrict')


class CLDataTypeSemantics(object):
    def result(self, ast):
        return ast

    def expr(self, ast):
        return ast

    def data_type(self, ast):
        return ast

    def is_pointer(self, ast):
        return ast

    def scalar_data_type(self, ast):
        return ast

    def vector_data_type(self, ast):
        return ast

    def user_data_type(self, ast):
        return ast

    def address_space_qualifier(self, ast):
        return ast

    def pre_asterisk_qualifiers(self, ast):
        return ast

    def post_asterisk_qualifiers(self, ast):
        return ast



import pyopencl as cl
import numpy as np
from ...utils import get_float_type_def, KernelInputDataManager
from ...cl_routines.base import CLRoutine
from ...load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CalculateModelEstimates(CLRoutine):

    def calculate(self, model, parameters):
        """Evaluate the model for every problem and every observation and return the estimates.

        This evaluates only the model and not the likelihood function of the model given the measured data.

        Args:
            model (AbstractModel): The model to evaluate.
            parameters (ndarray): The parameters to use in the evaluation of the model

        Returns:
            ndarray: Return per problem instance the evaluation per data point.
        """
        mot_float_dtype = np.float32
        if model.double_precision:
            mot_float_dtype = np.float64

        nmr_inst_per_problem = model.get_nmr_inst_per_problem()

        parameters = np.require(parameters, mot_float_dtype, requirements=['C', 'A', 'O'])

        nmr_problems = parameters.shape[0]
        evaluations = np.zeros((nmr_problems, nmr_inst_per_problem), dtype=mot_float_dtype, order='C')

        workers = self._create_workers(lambda cl_environment: _EvaluateModelWorker(
            cl_environment, self.get_compile_flags_list(model.double_precision), model, parameters, evaluations,
            mot_float_dtype))
        self.load_balancer.process(workers, nmr_problems)

        return evaluations


class _EvaluateModelWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, parameters, evaluations, mot_float_dtype):
        super(_EvaluateModelWorker, self).__init__(cl_environment)

        self._model = model
        self._data_info = self._model.get_kernel_data()
        self._data_struct_manager = KernelInputDataManager(self._data_info, mot_float_dtype)
        self._double_precision = model.double_precision
        self._evaluations = evaluations
        self._parameters = parameters

        self._all_buffers, self._evaluations_buffer = self._create_buffers()
        self._kernel = self._build_kernel(self._get_kernel_source(), compile_flags)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        kernel_func = self._kernel.get_estimates

        scalar_args = [None, None]
        scalar_args.extend(self._data_struct_manager.get_scalar_arg_dtypes())
        kernel_func.set_scalar_arg_dtypes(scalar_args)

        kernel_func(self._cl_queue, (int(nmr_problems), ), None,
                    *self._all_buffers, global_offset=(int(range_start),))
        self._enqueue_readout(self._evaluations_buffer, self._evaluations, range_start, range_end)

    def _create_buffers(self):
        evaluations_buffer = cl.Buffer(self._cl_context,
                                       cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                       hostbuf=self._evaluations)

        all_buffers = [cl.Buffer(self._cl_context,
                                 cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                 hostbuf=self._parameters),
                       evaluations_buffer]
        all_buffers.extend(self._data_struct_manager.get_kernel_inputs(self._cl_context, 1))
        return all_buffers, evaluations_buffer

    def _get_kernel_source(self):
        eval_function_info = self._model.get_model_eval_function()
        param_modifier = self._model.get_pre_eval_parameter_modifier()

        nmr_params = self._parameters.shape[1]

        kernel_param_names = ['global mot_float_type* restrict params',
                              'global mot_float_type* restrict estimates']
        kernel_param_names.extend(self._data_struct_manager.get_kernel_arguments())

        kernel_source = '''
            #define NMR_INST_PER_PROBLEM ''' + str(self._model.get_nmr_inst_per_problem()) + '''
        '''
        kernel_source += get_float_type_def(self._model.double_precision)
        kernel_source += self._data_struct_manager.get_struct_definition()
        kernel_source += eval_function_info.get_cl_code()
        kernel_source += param_modifier.get_cl_code()
        kernel_source += '''
            __kernel void get_estimates(
                ''' + ",\n".join(kernel_param_names) + '''
                ){
                    ulong gid = get_global_id(0);
                    mot_data_struct data = ''' + self._data_struct_manager.get_struct_init_string('gid') + ''';

                    mot_float_type x[''' + str(nmr_params) + '''];
                    for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                        x[i] = params[gid * ''' + str(nmr_params) + ''' + i];
                    }

                    global mot_float_type* result = estimates + gid * NMR_INST_PER_PROBLEM;
                    
                    ''' + param_modifier.get_cl_function_name() + '''(&data, x);
                    
                    for(uint i = 0; i < NMR_INST_PER_PROBLEM; i++){
                        result[i] = ''' + eval_function_info.get_cl_function_name() + '''(&data, x, i);
                    }
            }
        '''
        return kernel_source



import unittest
from scipy.stats import gamma
import numpy as np
from numpy.testing import assert_allclose

from mot.library_functions import GammaCDF
from mot.utils import cartesian


class test_GammaFunctions(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(test_GammaFunctions, self).__init__(*args, **kwargs)

    def test_gamma_cdf(self):
        test_params = self._gamma_cdf_params().astype(dtype=np.float32)

        python_results = self._calculate_python(test_params)
        cl_results = self._calculate_cl(test_params)

        assert_allclose(np.nan_to_num(python_results), np.nan_to_num(cl_results), atol=1e-5, rtol=1e-5)

    def _gamma_cdf_params(self):
        """Get some test params to test the Gamma CDF
        """
        shapes = np.arange(0.1, 5, 0.2)
        scales = np.arange(0.1, 5, 0.2)
        xs = np.arange(0.1, 10, 0.2)
        arr = cartesian([shapes, scales, xs])
        return arr

    def _calculate_cl(self, test_params):
        test_params = test_params.astype(np.float64)
        return GammaCDF().evaluate({
            'shape': test_params[:, 0],
            'scale': test_params[:, 1],
            'x': test_params[..., 2]})

    def _calculate_python(self, input_params):
        results = np.zeros(input_params.shape[0])

        for ind in range(input_params.shape[0]):
            shape = input_params[ind, 0]
            scale = input_params[ind, 1]
            x = input_params[ind, 2]

            results[ind] = gamma.cdf(x, shape, scale=scale)
        return results
